'''
深度学习基础工具
'''
# import sys
# import os
# import cv2
import numpy as np
# from scipy import misc
import matplotlib.pyplot as plt
# import re


# **********************************************
# **********  Result visualization  ************
# **********************************************
def visualize_feature_map(FM, outdir, num=1, name=None):
    '''
    Parse the input FeatureMap and do sample visualization
    Args:
        input: Tensor of feature map.
        outdir: Dir of ouput image saving.

    '''
    FM = np.array(FM)
    if len(FM.shape) == 5:  # [11HWC] -> [1HWC]
        FM = FM[0]
    if len(FM.shape) == 4:  # [1HWC] -> [HWC]
        FM = FM[0]

    h, w, c = FM.shape
    # 归一化 0-1
    tp_min, tp_max = np.min(FM), np.max(FM)
    if tp_min < 0:
        FM += abs(tp_min)
    FM /= (tp_max - tp_min)

    # Iterate filters
    for i in range(c):
        fig = plt.figure(figsize=(12, 12))
        axes = fig.add_subplot(111)
        img = FM[:, :, i]
        # Toning color
        axes.imshow(img, vmin=0, vmax=0.9, interpolation='bicubic', cmap='coolwarm')
        # Remove any labels from the axes
        # axes.set_xticks([])
        # axes.set_yticks([])

        # Save figure
        if name is None:
            # misc.imsave(outdir + '/%d_%.3d.png' % (num, i), FM[:, :, i])
            plt.savefig(outdir + '/%d_%.3d.png' % (num, i), dpi=60, bbox_inches='tight')
        else:
            # misc.imsave(outdir + '/%s_%.3d.png' % (name, i), FM[:, :, i])
            plt.savefig(outdir + '/%s_%.3d.png' % (name, i), dpi=60, bbox_inches='tight')
        plt.close(fig)


if __name__ == '__main__':
    # in_dir = '/home/tao/Data/VOCdevkit2007/VOC2007/small/train'
    # rename_files(in_dir)
    # mkdir_of_dataset('/home/tao/Data/XM')
    pass



# -*- coding:utf-8 -*-
'''
Files tools.

Version 1.0  2018-10-25 16:36:55
by QiJi Refence:
TODO:
1. xxx

'''
# import sys
import os
import re


# **********************************************
# ***************     ****************
# **********************************************
def rename_file(file_name, ifID=0, addstr=None, extension=None):
    '''Rename a file.
    Args:
        file_name: The name/path of file.
        ifID: 1 - only keep the number(ID) in old_name
            Carefully! if only keep ID, file_name can't be path.
        addstr: The addition str add between name and extension
        extension: Set the new extension(kind of image, such as: 'png').
    '''
    savename, extn = os.path.splitext(file_name)  # extn content '.'
    if ifID:
        # file_path = os.path.dirname(full_name)
        ID_nums = re.findall(r"\d+", savename)
        ID_str = str(ID_nums[0])
        for i in range(len(ID_nums)-1):
            ID_str += ('_'+(ID_nums[i+1]))
        savename = ID_str

    if addstr is not None:
        savename += '_' + addstr

    if extension is not None:
        extn = '.' + extension

    return savename + extn


def mkdir_nonexist(path):
    ''' Create a folder if it does not exist. That path can be return. '''
    if not os.path.exists(path):
        os.mkdir(path)
    return path


def mkdir_of_dataset(data_dir):
    '''
    Create folders of DL datasets according to the standard structure:
        ├── "dataset_name"(data_dir)
        |   ├── Log: log the train details
        |   ├── Model
        |       ├── checkpoints
        |   ├── SAVE: save some checkpoints and results
        |   ├── BackUp: backup some data or code
        |   ├── train
        |   ├── train_labels
        |   ├── val
        |   ├── val_labels
        |   ├── test
        |   ├── test_labels
    '''
    dir_list = ['/Log', '/Model', '/Model/checkpoints',
                '/SAVE', '/BackUp', '/train', '/train_labels',
                '/val', '/val_labels', '/test', '/test_labels']
    for a_dir in dir_list:
        mkdir_nonexist(data_dir+a_dir)
        print('make %s' % (a_dir))


def mkdir_of_classifyresult(out_dir, class_list):
    '''
    Make dir of (classification)reslut
    Args:
        out_dir: Dir of output
        class_list: list of class name
    '''
    for a_class in class_list:
        mkdir_nonexist(out_dir + "/" + a_class)


def filelist(floder_dir, ifPath=False, extension=None):
    '''
    Get names(or whole path) of all files(with specify extension)
    in the floder_dir and return as a list.

    Args:
        floder_dir: The dir of the floder_dir.
        ifPath:
            True - Return whole path of files.
            False - Only return name of files.(Defualt)
        extension: Specify extension to only get that kind of file names.

    Returns:
        namelist: Name(or path) list of all files(with specify extension)
    '''
    namelist = sorted(os.listdir(floder_dir))

    if ifPath:
        for i in range(len(namelist)):
            namelist[i] = os.path.join(floder_dir, namelist[i])

    if extension is not None:
        n = len(namelist)-1  # orignal len of namelist
        for i in range(len(namelist)):
            if not namelist[n-i].endswith(extension):
                namelist.remove(namelist[n-i])  # discard the files with other extension

    return namelist


def filepath_to_name(full_name, extension=False):
    '''
    Takes an absolute file path and returns the name of the file with(out) the extension.
    '''
    file_name = os.path.basename(full_name)
    if not extension:  # if False then discard extension
        file_name = os.path.splitext(file_name)[0]
    return file_name


# **********************************************
# ************ Main functions ******************
# **********************************************
def rename_files(input_dir, out_dir=None, num=1):
    ''' Rename all the files in a floder by number.
    Args:
        input_dir: Original folder directory
        out_dir: Renamed files output directory(optional)
        num: The starting number of renamed files(default=1)
    '''
    # Log the rename record
    folder_dir = os.path.dirname(input_dir)
    folder_name = os.path.basename(input_dir)
    target = open(folder_dir+"/ReName Log_%s.txt" % folder_name, 'w')

    file_names = sorted(os.listdir(input_dir))
    extension = os.path.splitext(file_names[0])[1]  # Plan A
    if out_dir is None:
        out_dir = input_dir
    for name in file_names:
        newname = ("%.5d" % (num))+extension  # Plan A
        # newname = rename_file(name, ifID=1)  # Plan B
        # TODO: 文件名中文乱码，下同
        os.rename(input_dir+'/'+name, input_dir+'/'+newname)
        target.write(name + '\tTo\t' + newname + '\n')
        num += 1

    target.close()
    print("Finish rename.")


def main():
    pass


if __name__ == '__main__':
    # main()
    # mkdir_of_dataset('/home/tao/Data/RBDD')
    pass



'''
Tools collection of Data(np.ndarray) processing.

Version 1.0  2018-04-03 15:44:13 by QiJi
Version 1.5  2018-04-07 22:34:53 by QiJi
Version 2.0  2018-04-11 09:43:47 by QiJi
    Add MyData class, rename this model names and some functions name
Version 3.0  2018-10-25 16:59:23 by QiJi
'''
import os
import sys
import cv2
import numpy as np
from tqdm import tqdm
from .fileop import filelist


def get_label_info(file_path):
    """
    Retrieve the class names and label values for the selected dataset.
    Must be in CSV or Txt format!

    Args:
        file_path: The file path of the class dictionairy

    Returns:
        class_names: A list of class names.
        label values: A list per class's color values. [[0,0,0], [255,255,255], ]
    """
    import csv
    filename, exten = os.path.splitext(file_path)
    if not (exten == ".csv" or exten == ".txt"):
        return ValueError("File is not a CSV or TxT!")

    class_names, label_values = [], []
    with open(file_path, 'r') as f:
        file_reader = csv.reader(f, delimiter=',')
        row = next(file_reader)  # skip one line
        for row in file_reader:
            if row != []:
                class_names.append(row[0])
                label_values.append([int(row[1]), int(row[2]), int(row[3])])
    return class_names, label_values


def compute_class_weights(labels_dir, label_values):
    '''
    Arguments:
        labels_dir(list): Directory where the image segmentation labels are
        num_classes(int): the number of classes of pixels in all images

    Returns:
        class_weights(list): a list of class weights where each index represents
            each class label and the element is the class weight for that label.

    '''
    # TODO: need Test and Debug
    image_files = filelist(labels_dir, ifPath=True, extension='.png')
    num_classes = len(label_values)
    class_pixels = np.zeros(num_classes)
    total_pixels = 0.0

    for n in range(len(image_files)):
        image = cv2.imread(image_files[n], -1)
        for index, colour in enumerate(label_values):
            class_map = np.all(np.equal(image, colour), axis=-1)
            class_map = class_map.astype(np.float32)
            class_pixels[index] += np.sum(class_map)

        print("\rProcessing image: " + str(n) + " / " + str(len(image_files)), end="")
        sys.stdout.flush()

    total_pixels = float(np.sum(class_pixels))
    index_to_delete = np.argwhere(class_pixels == 0.0)
    class_pixels = np.delete(class_pixels, index_to_delete)
    class_weights = total_pixels / class_pixels
    class_weights = class_weights / np.sum(class_weights)

    return class_weights


# **********************************************
# *********** Data Post-treatment **************
# **********************************************
def bwmorph(image, kernel_1=None, kernel_2=None):
    '''
    Do BW morphologic pocessing in label.

    Args:
        image: Tnput image, better be [HW].(also support [HW1] and [HWC])
        kernel_1: The kernel for closses

    Returns:
        image: [HW] label

    Notice: If input label is a multiply channel img, will only keep
        its first channel, then do morphologic pocessing.
    '''
    kernel_0 = np.array([[0, 0, 1, 0, 0],
                         [0, 1, 1, 1, 0],
                         [1, 1, 1, 1, 1],
                         [0, 1, 1, 1, 0],
                         [0, 0, 1, 0, 0]],
                        dtype=np.uint8)
    if kernel_1 is None:
        kernel_1 = kernel_0
    if kernel_2 is None:
        kernel_2 = kernel_0

    if len(image.shape) > 2:
        if image.shape[2] == 3:
            image = image[:, :, 0]
        elif image.shape[2] == 1:
            image = image[:, :, 0]
        else:
            ValueError("What a fucking image U choose?")

    # 闭运算-用[5x5](默认)kernel连接细小断裂
    image = cv2.erode(cv2.dilate(image, kernel_1), kernel_1)

    # 开运算-用[5x5]kernel去除细微噪点
    # kernel = np.ones([5, 5], dtype=np.uint8)
    image = cv2.dilate(cv2.erode(image, kernel_2), kernel_2)

    return image


def vote_combine(label, crop_params, crop_info, mode, scale=False):
    '''
    Combine small scale predicted label into a big one,
    for 1.classification or 2.semantic segmantation.
    Args:
        label: One_hot label(may be tensor). 1.[NC]; 2.[NHWC]
        crop_params: [sub_h, sub_w, crop_stride]
        crop_info: [rows, cols]
            rows: Num of images in the row direction.
            cols: Num of images in the col direction.
        mode: 1-classification, 2-semantic segmantation.
        scale: If do adaptive normalize.
    Returns:
        out_label: [HWC] one hot array, uint8.
    '''
    rows, cols = crop_info
    h, w, stride = crop_params
    label = np.array(label)  # Tensor -> Array
    if mode == 1:
        # 1. For classification
        if len(label.shape) == 3:  # list转array后会多出一维
            label = np.squeeze(label)
    elif mode == 2:
        # 2. For semantic segmantation
        if len(label.shape) == 5:  # in case of label is [1NHWC]
            label = label[0]
    else:
        return ValueError('Incorrect mode!')
    out_h, out_w = (rows-1)*stride+h, (cols-1)*stride+w
    out_label = np.zeros([out_h, out_w, label.shape[-1]], dtype=label.dtype)  # [HWC]

    y = 0  # y=h=row
    for i in range(rows):
        x = 0  # x=w=col
        for j in range(cols):
            out_label[y:y+h, x:x+w] += label[i*cols+j]  # 此处融合用加法
            x += stride
        y += stride

    # 自适应归一化
    if scale:
        m = np.max(out_label, axis=-1, keepdims=True).repeat(2, -1)
        n = np.min(out_label, axis=-1, keepdims=True).repeat(2, -1)
        out_label = out_label / (m - n)
    return out_label  # np.uint8(out_label)


def modefilter(label):
    '''找到label matrix中所有的无类别值(0)，用8连通区的众数给其赋值'''
    # TODO:How to solve the problem of boundary.
    newlabel = label.copy()
    indlist = np.argwhere(label == 0)
    for [r, c] in indlist:  # ind - [r, c]
        try:
            connected = label[r-1:r+1, c-1:c+1]
            mode = np.argmax(np.bincount(connected))
        except Exception:
            mode = 0
        newlabel[r, c] = mode
    return newlabel


# **********************************************
# ************ Main functions ******************
# **********************************************
def data_statistics():
    ''' Count (pixel-level) mean and variance of dataset. '''
    dir_list = [
        '/home/tao/Data/RBDD/train',
    ]  # list of data folder dir.

    mean = np.array([0, 0, 0], np.float32)  # BGR
    # std = np.array([0, 0, 0],  np.float32)  # BGR
    total_image = 0
    # Statisify mean and std
    for folder in dir_list:
        image_paths = filelist(folder, ifPath=True)
        for path in tqdm(image_paths):
            img = cv2.imread(path, cv2.IMREAD_COLOR)  # BGR
            img = np.float32(img) / 255.0
            # images.append(np.expand_dims(img))
            mean += np.mean(np.mean(img, 0), 0)
            total_image += 1
    mean /= total_image  # get pixel-level BGR mean
    # Statisify std
    # for folder in dir_list:
    #     image_paths = filelist(folder, ifPath=True)
    #     for path in tqdm(image_paths):
    #         img = cv2.imread(path, cv2.IMREAD_COLOR)  # BGR
    #         img = np.float32(img) / 255.0
    #         X = img-mean
    #         X = X*X
    #         std += np.sum(np.sum(X, axis=0), axis=0)
    # std /= total_image
    print('mean=', mean)
    # print('std=', std)
    print('Careful: BGR!')
    # return mean, std


def spilt_dataset(in_floders, out_floders, rate=0.025, max_num=None):
    '''Spilt images(maybe with labels)
    Args:
        in_floders - list of input floders
        out_floders - list of output floders
            for example:
                in_floders = [
                    'dataset_dir/train',
                    'dataset_dir/train_labels'
                ]
                out_floders = [
                    'dataset_dir/val',
                    'dataset_dir/val_labels'
                ]
    '''
    from shutil import move
    # from sklearn.model_selection import train_test_split

    name_list = []
    for d in in_floders:
        names = filelist(d)
        name_list.append(names)
    file_num = len(name_list[0])  # 文件总数
    max_num = file_num if max_num is None else max_num

    # 按比例抽取
    index = np.random.permutation(file_num)
    index = index[:int(file_num*rate)]

    for d, names in zip(range(len(in_floders)), name_list):
        print(in_floders[d])
        cnt = 0
        for i in index:
            # 针对文件的操作
            # os.remove(d+'/'+names[i])
            move(in_floders[d]+'/'+names[i], out_floders[d]+'/'+names[i])
            cnt += 1
            if cnt >= max_num:
                break  # if set max_num, prioritize the max_num

    print('Finished!')


def MorphologicPostTreate(input_dir, output_dir):
    '''
    形态学后处理
    '''
    image_names = filelist(input_dir)
    kernel = np.ones([7, 7], dtype=np.uint8)
    kernel[0, 0], kernel[-1, -1], kernel[0, -1], kernel[-1, 0] = 0, 0, 0, 0
    # print(kernel)

    for i in tqdm(range(len(image_names))):
        img = cv2.imread(input_dir + "/" + image_names[i], -1)
        img = bwmorph(img, kernel_1=kernel)  # 输出[HW]label
        img = np.repeat(np.expand_dims(img, axis=2), 3, axis=2)
        cv2.imwrite(output_dir + "/" + image_names[i], img)


def main():
    pass


if __name__ == '__main__':
    # main()
    data_statistics()
    # spilt_dataset(max_num=1000)
    pass



'''
Tools collection of Data(np.ndarray) processing.

Note:
    1. cv2.imwrite(image_name, image, [1, 100]),
        [1, 100] mean set [cv2.IMWRITE_JPEG_QUALITY, 100]

Version 1.0  2018-04-03 15:44:13 by QiJi
Version 1.5  2018-04-07 22:34:53 by QiJi
Version 2.0  2018-10-25 16:59:23 by QiJi
'''
import os

# import sys
import cv2
import h5py
import numpy as np
from tqdm import tqdm

from .dldata import get_label_info
from .fileop import filelist, rename_file


# **********************************************
# ************ Image basic tools ***************
# **********************************************
def load_image(path, RGB=0):
    '''Unchange mode to load image(and may be convert it to RGB)
    Args:
        path: The whole path of image.
        RGB: 0-BGR(opencv deafault); 1-RGB(use cvtColor convert it to RGB)
    Returns:
        image: image
    '''
    image = cv2.imread(path, -1)  # BGR
    if RGB:
        cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


def convert_image_type(image_dir, new_type, old_type=None):
    '''Covert image's type(may be specify the old type).
    Args:
        new_type: The target type of image conversion(such as: 'png', 'jpg').
    '''
    image_names = filelist(image_dir, True, old_type)
    for name in tqdm(image_names):
        img = cv2.imread(name, 1)  # 默认BGR模式读，适应Tiff的标签图
        os.remove(name)
        name = os.path.splitext(name)[0]+'.'+new_type
        cv2.imwrite(name, img, [1, 100])


def one_hot_1(label, class_num):
    '''One hot code the label, not support RBG label.
    Args:
        label: a [HW] or [HW1] array
        class_num: num of classes
    Returns:
        one_hot: a 3D array of one_hot label (dtype=np.uint8), C=num of classes
    '''
    one_hot = np.zeros([label.shape[0], label.shape[1], class_num], dtype=np.uint8)
    for i in range(class_num):
        one_hot[:, :, i] = (label == i)  # TODO: need test [HW1]
    return one_hot


def one_hot_2(label, label_values):
    """
    One hot code the RBG label by replacing each pixel value with a vector of length num_classes
    Note: If class_dict is RGB, the label must be RGB(not BGR).
    Args:
        label: a [HWC] array, and C=3(be carefull about RGB or BGR)
        label_values: A list per class's color values.  [[0,0,0], [255,255,255], ]
    Returns:
        one_hot: one_hot label, C=num of classes
    """
    semantic_map = []
    for colour in label_values:
        equality = np.equal(label, colour)
        class_map = np.all(equality, axis=-1)
        semantic_map.append(class_map)
    semantic_map = np.stack(semantic_map, axis=-1).astype(np.uint8)

    return semantic_map


def one_hot_3(label, class_num):
    '''One hot code the label, classification result.
    Args:
        label: a [1] or [N1] array
        class_num: num of classes
    Returns:
        one_hot: a [NC] array of one_hot label, C=num of classes
    '''
    one_hot = np.zeros([label.shape[0], class_num], dtype=label.dtype)
    for i in range(class_num):
        one_hot[:, i] = (label == i)
    return one_hot


def class_label(label, label_values):
    '''
    Convert RGB label to 2D [HW] array, each pixel value is the classified class key.
    '''
    semantic_map = np.zeros(label.shape[:2], label.dtype)
    for i in range(len(label_values)):
        equality = np.equal(label, label_values[i])
        class_map = np.all(equality, axis=-1)
        semantic_map[class_map] = i
    return semantic_map


def reverse_one_hot(one_hot):
    '''Transform a 3D array in one-hot format (depth is num_classes),
    to a 2D array, each pixel value is the classified class key.
    '''
    return np.argmax(one_hot, axis=2)  # set axis=2 to limit the input is 3D


def colour_code_label(label, label_values, add_image=None, save_path=None):
    '''
    Given a [HW] array of class keys(or one hot[HWC]), colour code the label;
    also can weight the colour coded label and image, maybe save the final result.

    Args:
        label: single channel array where each value represents the class key.
        label_values: A list per class's color values. [[0,0,0], [255,255,255], ]
    Returns:
        Colour coded label or just save image return none.
    '''
    label, colour_codes = np.array(label), np.array(label_values)
    if len(label) == 3:
        label = np.argmax(label, axis=2)  # [HWC] -> [HW]
    color_label = colour_codes[label.astype(int)]  # TODO:此处直接uint8有错误
    color_label = color_label.astype(np.uint8)

    if add_image is not None:
        if add_image.shape != color_label.shape:
            cv2.resize(color_label, (add_image.shape[1], add_image.shape[0]),
                       interpolation=cv2.INTER_NEAREST)
        add_image = cv2.addWeighted(add_image, 0.7, color_label, 0.3, 0)
        if save_path is None:
            return color_label, add_image

    if save_path is not None:
        cv2.imwrite(save_path, color_label, [1, 100])
        if add_image is not None:
            cv2.imwrite(rename_file(save_path, addstr='mask'), add_image, [1, 100])
        return  # no need to return label if saved

    return color_label


def mask_img(image, label, mask_value=[0, 0, 0]):
    '''Mask image with a specified value of label.
    Note: mask_value is BGR list
    '''
    equality = np.equal(label, mask_value)
    mask = np.all(equality, axis=-1)
    image[mask] = [0, 0, 0]
    return image


# **********************************************
# ******** Common data Pre-treatment ***********
# **********************************************
def slide_crop(image, crop_params=[256, 256, 128], ifPad=True, ifbatch=False):
    '''
    Slide crop image into small piece, return list of sub_image or a [NHWC] array.
    Args:
        crop_params: [H, W, stride] (default=[256, 256, 128])
        ifbatch: Ture-return [NHWC] array; False-list of image.
    Return:
        size: [y, x]
            y: Num of images in the row direction.
            x: Num of images in the col direction.
    Carefully: If procided clipping parameters cannot completely crop the entire image,
        then it cannot be restored to original size during the recovery process.
    '''
    crop_h, crop_w, stride = crop_params
    h, w = image.shape[:2]

    h_drop, w_drop = (h-crop_h) % stride, (w-crop_w) % stride
    if h_drop or w_drop:
        # Cannot completely slide-crop the target image
        if ifPad:
            image, _ = pad_img(image, crop_h, crop_w, stride, stride)
            h, w = image.shape[:2]  # Update image size
        else:
            image = image[:-h_drop, :-w_drop]

    image_list = []
    y = 0  # y:h(row)
    for i in range((h-crop_h)//stride + 1):
        x = 0  # x:w(col)
        for j in range((w-crop_w)//stride + 1):
            tmp_img = image[y:y+crop_h, x:x+crop_w].copy()
            if ifbatch:
                tmp_img = np.expand_dims(tmp_img, axis=0)
            image_list.append(tmp_img)
            x += stride
        y += stride
    size = [i+1, j+1]

    if ifbatch:
        batch_size = len(image_list)
        if batch_size == 1:
            return image_list[0], size
        else:
            image_bantch = np.squeeze(np.stack(image_list, axis=1))
            return image_bantch, size

    return image_list, size


def pad_img(image, kh, kw, sh, sw):
    '''Pad image according kernel size and stride.
    Args:
        image - array
        kh, kw - kernel height, kernel width
        sh, sw - height directional stride, width directional stride.
    '''
    h, w = image.shape[:2]
    d = len(image.shape)
    pad_h, pad_w = sh - (h-kh) % sh, sw - (w-kw) % sw
    pad_h = (pad_h//2, pad_h//2+1) if pad_h % 2 else (pad_h//2, pad_h//2)
    pad_w = (pad_w//2, pad_w//2+1) if pad_w % 2 else (pad_w//2, pad_w//2)
    pad_params = (pad_h, pad_w) if d == 2 else (pad_h, pad_w, (0, 0))
    return np.pad(image, pad_params, mode='constant'), pad_params


def resized_crop(image, i, j, h, w, size, interpolation=cv2.INTER_LINEAR):
    '''Crop the given PIL Image and resize it to desired size.
    Args:
        i: Upper pixel coordinate.
        j: Left pixel coordinate.
        h: Height of the cropped image.
        w: Width of the cropped image.
        size: (Height, Width) must be tuple
    '''
    image = image[i:i+h, j:j+w]
    image = cv2.resize(image, size[::-1], interpolation)
    return image


def random_crop(image, crop_height=256, crop_width=256):
    '''
    Crop image(label) randomly
    '''
    if (crop_height < image.shape[0] and crop_width < image.shape[1]):
        x = np.random.randint(0, image.shape[0] - crop_height)  # row
        y = np.random.randint(0, image.shape[1] - crop_width)  # column
        return image[x:x + crop_height, y:y + crop_width]

    elif (crop_height == image.shape[0] and crop_width == image.shape[1]):
        return image

    else:
        raise Exception('Crop size > image.shape')


def random_crop_2(image, label, crop_height=256, crop_width=256):
    '''
    Crop image and label randomly

    '''
    if image.shape[0] != label.shape[0] or image.shape[1] != label.shape[1]:
        raise Exception('Image and label must have the same shape')
    if (crop_height < image.shape[0] and crop_width < image.shape[1]):
        x = np.random.randint(0, image.shape[0] - crop_height)  # row
        y = np.random.randint(0, image.shape[1] - crop_width)  # column
        # label maybe multi-channel[H,W,C] or one-channel [H,W]
        return image[x:x + crop_height, y:y + crop_width], label[
                x:x + crop_height, y:y + crop_width]
    elif (crop_height == image.shape[0] and crop_width == image.shape[1]):
        return image, label
    else:
        raise Exception('Crop size > image.shape')


def random_crop_3(image, label, crop_h=256, crop_w=256, num_classes=0):
    '''
    Crop image(label) randomly for serval times(now four times most),
    and return the one in which per class RATIO most balancing.
    Don't support RGB label now.
    Args:
        label: 1. [HW]; 2. [HW1]; 3. one_hot-[HWC]; 4. rgb-[HWC]-目前不支持rbg!!!
            Note: 1&2 class_num need to be specified!!!
        num_classes: default=0, only one_hot label don't have to specify.
    Returns:
        image: croped image
        label: croped laebl
    '''
    if (image.shape[0] != label.shape[0]) or (image.shape[1] != label.shape[1]):
        raise Exception('Image and label must have the same dimensions!')

    if (crop_w < image.shape[1]) and (crop_h < image.shape[0]):
        x, y = [0, 0, 0, 0], [0, 0, 0, 0]
        rate = [1., 1., 1., 1.]  # store min RATIO
        sum_area = crop_h * crop_w

        if len(label.shape) == 2 or label.shape[-1] == 1:  # [HW] or [HW1]
            for i in range(4):  # try four times and choose the max rate's x,y
                x[i] = np.random.randint(0, image.shape[1] - crop_w)  # W
                y[i] = np.random.randint(0, image.shape[0] - crop_h)  # H
                tmp_label = label[y[i]:y[i] + crop_h, x[i]:x[i] + crop_w].crop()

                for j in range(num_classes):
                    indArr = [tmp_label == j]
                    tmp_rate = np.sum(indArr) / sum_area
                    if tmp_rate == 0:
                        rate[i] = tmp_rate
                        break
                    elif tmp_rate < rate[i]:
                        rate[i] = tmp_rate

        else:  # [HWC] - only support one_hot label now
            for i in range(4):
                x[i] = np.random.randint(0, image.shape[1] - crop_w)  # W
                y[i] = np.random.randint(0, image.shape[0] - crop_h)  # H
                tmp_label = label[y[i]:y[i] + crop_h, x[i]:x[i] + crop_w, :].crop()

                # For one_hot label
                for j in range(tmp_label.shape[2]):  # traverse all channel-class
                    indArr = tmp_label[:, :, j]
                    tmp_rate = np.sum(indArr) / sum_area
                    if tmp_rate == 0:
                        rate[i] = tmp_rate
                        break
                    elif tmp_rate < rate[i]:
                        rate[i] = tmp_rate

        ind = rate.index(max(rate))  # choose the max RATIO area
        x, y = x[ind], y[ind]

        try:
            label = label[y:y + crop_h, x:x + crop_w]
        except Exception:
            label = label[y:y + crop_h, x:x + crop_w, :]

        image = image[y:y + crop_h, x:x + crop_w, :]
        return image, label

    elif (crop_w == image.shape[1] and crop_h == image.shape[0]):
        return image, label

    else:
        raise Exception('Crop shape exceeds image dimensions!')


def resize_img(image, new_size=[256]):
    '''
    Resize the image into a new shape (keep aspect ratio,
    let the shortest side equal to the target heigth/width,
    then crop it into the targe height and width).
    Args:
        new_size - a tuple may have only one element(H=W) or two elements(H, W).
    '''
    if len(new_size) == 2:
        H, W = new_size
        image = cv2.resize(image, (W, H))
    elif len(new_size) == 1:
        h, w = image.shape[:2]
        if w > h:  # if image width > image height
            H, W = new_size[0], int(w*new_size[0]/h)
            st = int((W-H)/2)
            image = cv2.resize(image, (W, H))[:, st:st+H]
        else:  # if image height > image width
            H, W = int(w*new_size[0]/h), new_size[0]
            st = int((H-W)/2)
            image = cv2.resize(image, (W, H))[st:st+H, :]
    else:
        ValueError('Incorrect new_size!')

    return image


def unify_imgsize(img_dir, size=(256, 256), interpolation=cv2.INTER_NEAREST):
    '''
    Unify image size.
    Args:
        size: Uniform size(must be tuple)
        interpolation: Interpolation method of zoom image
    '''
    num = 1
    image_names = sorted(os.listdir(img_dir))
    for name in tqdm(image_names):
        img = cv2.imread(img_dir+'/'+name, -1)
        if img.shape[:2] != size:
            img = cv2.resize(img, size[::-1], interpolation)
        cv2.imwrite(img_dir+'/'+name, img, [1, 100])
        num += 1


def label_pretreat(label_dir, label_values):
    '''
    Pre-treat the orignal RGB label, to fix the apparent bug.
    (Actually it eliminate the wrong colors that are not in class_dict)
    By the way, unify the dtype of label imgs to png.
    '''
    l_names = filelist(label_dir, ifPath=True)
    for name in tqdm(l_names):
        label = cv2.imread(name, 1)  # read in RGB
        os.remove(name)
        name = os.path.splitext(name)[0] + '.png'
        new_label = np.zeros(label.shape, label.dtype)
        # Fix the color(stand for a class) by class-info
        for color in label_values:
            equality = np.equal(label, color)
            ind_mat = np.all(equality, axis=-1)
            new_label[ind_mat] = color  # this color list can be customized
        cv2.imwrite(name, new_label)  # new_label-BGR(according to class_dict)


# **********************************************
# ********** Commen data augment ***************
# **********************************************
def gamma_transform(img, gamma):
    gamma_table = [np.power(x / 255.0, gamma) * 255.0 for x in range(256)]
    gamma_table = np.round(np.array(gamma_table)).astype(np.uint8)
    return cv2.LUT(img, gamma_table)


def random_gamma_transform(img, gamma_vari):
    log_gamma_vari = np.log(gamma_vari)
    alpha = np.random.uniform(-log_gamma_vari, log_gamma_vari)
    gamma = np.exp(alpha)
    return gamma_transform(img, gamma)


def rotate(xb, yb, angle):
    '''rotate xb, yb'''
    h, w = xb.shape[0], xb.shape[1]
    M_rotate = cv2.getRotationMatrix2D((w // 2, h // 2), angle, 1)
    xb = cv2.warpAffine(xb, M_rotate, (w, h))
    yb = cv2.warpAffine(yb, M_rotate, (w, h))
    return xb, yb


def blur(img):
    img = cv2.blur(img, (3, 3))
    return img


def add_noise(image, gamma=0.01):
    ''' 添加点噪声 '''
    num = image.shape[0] * image.shape[1] * gamma
    for i in range(num):  # 100 - control the num of piexls to add noise
        temp_x = np.random.randint(0, image.shape[0])
        temp_y = np.random.randint(0, image.shape[1])
        image[temp_x][temp_y] = np.random.randint(0, 255)

    return image


def data_augment(image, label=None, crop_size=None, zoom_size=None, if_flip=True, if_roate=True):
    '''
    Simple data augment function to random crop,flip and roate image and label.
    Args:
        image: array of image, [HWC] or [HW1] or [HW].
        label: array of label, [HWC] or [HW1] or [HW].
        crop_size: [h, w]
        zoom_size: [h, w]
        if_flip: if flip the image(and label), defualt = True
        if_roate: if roate the image(and label), defualt = True

    Returns:
        image, label: processed image and label
    '''
    if label is None:
        if crop_size is not None:
            image = random_crop(image, crop_size[0], crop_size[1])
        if zoom_size is not None:
            image = cv2.resize(
                image, (zoom_size[1], zoom_size[0]), interpolation=cv2.INTER_LINEAR)
        # Flip
        if if_flip and np.random.randint(0, 2):
            image = cv2.flip(image, 1)
        if if_flip and np.random.randint(0, 2):
            image = cv2.flip(image, 0)
        # Roate
        iflag = np.random.randint(0, 2)
        if if_roate and iflag:
            angle_table = [90, 180, 270]
            angle = angle_table[np.random.randint(0, 3)]
            h, w = image.shape[0], image.shape[1]
            M_rotate = cv2.getRotationMatrix2D((w // 2, h // 2), angle, 1)
            image = cv2.warpAffine(image, M_rotate, (w, h))
        return image
    else:
        if crop_size is not None:
            image, label = random_crop_2(image, label, crop_size[0], crop_size[1])
        if zoom_size is not None:
            image = cv2.resize(image, (zoom_size[1], zoom_size[0]), cv2.INTER_LINEAR)
            label = cv2.resize(label, (zoom_size[1], zoom_size[0]), cv2.INTER_NEAREST)
        # Flip
        if if_flip and np.random.randint(0, 2):
            image, label = cv2.flip(image, 1), cv2.flip(label, 1)
        if if_flip and np.random.randint(0, 2):
            image, label = cv2.flip(image, 0), cv2.flip(label, 0)
        # Roate
        iflag = np.random.randint(0, 2)
        if if_roate and iflag:
            angle_table = [90, 180, 270]
            angle = angle_table[np.random.randint(0, 3)]
            image, label = rotate(image, label, angle)
        return image, label


# **********************************************
# ************** HDF5 tools ********************
# **********************************************
def label2hdf5(input_dir, output_path, label_values=None, class_num=None):
    """
    Transform a label image into one-hot format (depth is num_classes),
    and save them all in a dataset(key:"labels") of a h5 flie.

    Args:
        input_dir: The dir of label images [HW] or [HWC](BGR)
        output_path: The whole path of the output h5 flie.
        label_values: A dictionairy of class--> pixel values, defualt=None
        class_num: Num of classes, defualt=None
            Note:
                1. If class_num is specified, input labels is [HW](imread in gray mode),
                    call one_hot_1
                2. If label_values is specified, input labels is [HWC]BGR, call one_hot_2;

    Returns
        None, but will create a 4d dataset store all one hot labels(in a h5 file).

    Notice:
        the dtype of the dataset is uint8(np.uint8).
        the key of labels dataset is "label".

    """
    print("*******Execute label2hdf5()*******")
    input_names = sorted(os.listdir(input_dir))  # only the name of labels
    # Carefull!! Since the os.listdir() return a list in atribute order, sort it

    # Read first of them to get shape
    tmp_input = cv2.imread(input_dir + "/" + input_names[0], -1)
    size = [0, 0, 0, 0]  # NHWC
    size[0] = len(input_names)  # get N
    size[1], size[2] = tmp_input.shape[0], tmp_input.shape[1]  # get H,W

    if class_num is not None:  # [HW], class_num must not None
        size[3] = class_num
    elif label_values is not None:  # [HWC]-BGR
        size[3] = size[3] = len(label_values)
    else:
        ValueError("label_values or class_num must specify one.")

    # Open(or create) a hdf5 flie to save all the labels
    f = h5py.File(output_path, 'a')
    dataset = f.create_dataset("label", shape=size, dtype=np.uint8)  # create a 4d dataset
    # Treat per label in the list
    if class_num is not None:  # [HW], class_num must not None
        for i in tqdm(range(len(input_names))):
            label = cv2.imread(input_dir + "/" + input_names[i], 0)
            dataset[i, :] = one_hot_1(label, class_num)  # return uint8 array
    elif label_values is not None:  # [HWC]-RGB
        target = open(os.path.dirname(output_path) + "/train_gt.txt", 'w')
        for i in tqdm(range(len(input_names))):
            label = cv2.imread(input_dir + "/" + input_names[i], -1)
            dataset[i, :] = one_hot_2(label, label_values)  # return uint8 array
            target.write(input_names[i]+'\n')
    print(dataset.shape, dataset.dtype)
    print("*******Finish label2hdf5()*******")
    f.close()


def img2hdf5(input_dir, output_path):
    """
    Save images and their names into h5 file

    Args:
        input_dir: The dir of images.
        output_path: The whole path of the output h5 flie.

    Returns:
        None.
        But will create a h5 file with a 4d dataset store all images(RGB),
            and a 1d dataset store all their names.
    Notice:
        the dtype of the dataset is uint8(np.uint8).
        the key of image dataset is "image".(RGB)
        the key of image name dataset is "name"

    """
    print("*******Execute img2hdf5()*******")
    input_names = sorted(os.listdir(input_dir))  # only the name of images
    # Carefull!! Since the os.listdir() return a list in atribute order, sort it
    names = []  # 将文件名转utf8编码，才能存入hdf5的dataset中
    for j in input_names:
        names.append(j.encode('utf8'))

    # Read first of them to get shape
    tmp_input = load_image(input_dir + "/" + input_names[0], 1)
    size = [0, 0, 0, 0]  # NHWC
    size[0] = len(input_names)  # get N
    size[1:] = tmp_input.shape  # get HWC

    # Open(or create) a hdf5 flie to save all the images and their names
    f = h5py.File(output_path, 'w')
    f.create_dataset("name", data=names)  # 创建name数据集存储images 的utf8编码的文件名
    # Create an empty 4d dataset(image) to store images
    dataset = f.create_dataset("image", shape=size, dtype=np.uint8)
    # Treat per img in the list
    target = open(os.path.dirname(output_path) + "/train.txt", 'w')
    for i in tqdm(range(len(input_names))):
        input_path = input_dir + "/" + input_names[i]  # the whole path
        img = load_image(input_path, 1)
        # Extension pocessing
        # for c in range(img.shape[-1]):
        #     img[:, :, c] = cv2.equalizeHist(img[:, :, c])
        dataset[i, :] = img
        target.write(input_names[i]+'\n')

    print(dataset.shape, dataset.dtype)
    print("*******Finish img2hdf5()*******")
    f.close()


def img2hdf5_2(input_dir, output_path, newsize=None):
    """
    For classified, save images and their names into h5 file,
    meanwhile make their labels[NC].

    Args:
        input_dir: The dir of images(in several folders).
        output_path: The whole path of the output h5 flie.

    Returns:
        None.
        But will create a h5 file with a 4d dataset store all images(RGB),
            and a 1d dataset store all their names.
    Notice:
        the dtype of the dataset is uint8(np.uint8).(RGB)
        the key of image dataset is "image".
        the key of label dataset is "label".
        the key of image name dataset is "name"
        the key of class name dataset is "class"

    """
    print("*******Execute img2hdf5()*******")
    # Fisrt travals get total image num and all class names and image names
    image_names = []  # all image names
    folder_names = sorted(os.listdir(input_dir))  # also name of classes
    for i, f in enumerate(folder_names):
        file_names = sorted(os.listdir(input_dir+'/'+f))  # name of images
        image_names += file_names

    utf_names = [name.encode('utf8') for name in image_names]  # 转utf8存入
    utf_classNames = [name.encode('utf8') for name in folder_names]

    # Read first of them to get shape
    tmp_input = load_image(os.path.join(input_dir, folder_names[0], image_names[0]), 1)
    img_size = [0, 0, 0, 0]  # NHWC(channel), size of image dataset
    label_size = [0, 0]  # NC(classnum), size of label dataset
    img_size[0] = label_size[0] = len(image_names)  # get N
    img_size[1:] = tmp_input.shape  # get HWC si.extend(ze
    label_size[1] = len(folder_names)  # get class num

    # Open(or create) a hdf5 flie to save all the images and their names
    f = h5py.File(output_path, 'w')
    f.create_dataset("name", data=utf_names)  # 创建name数据集存class_dict储images 的utf8编码的文件名
    f.create_dataset('class', data=utf_classNames)
    # Create an empty 4d dataset(image) to store images
    imageSet = f.create_dataset("image", shape=img_size, dtype=np.uint8)
    labelSet = f.create_dataset("label", shape=label_size, dtype=np.uint8)
    # Save per img.extend(
    i = 0
    for c, fo in tqdm(enumerate(folder_names)):
        file_names = sorted(os.listdir(input_dir+'/'+fo))
        for fi in file_names:
            img = load_image(os.path.join(input_dir, fo, fi), 1)
            if newsize is not None:
                img = cv2.resize(img, (newsize[1], newsize[0]))
            imageSet[i, :] = img
            labelSet[i, c] = 1
            i += 1

    # print(imageSet.shape, imageSet.dtype)
    print("*******Finish img2hdf5()*******")
    f.close()


# **********************************************
# ************ Main functions ******************
# **********************************************
def main_of_hdf5(img_dir, label_dir, class_dict_path, h5_path):
    ''' Cover images and labels to hdf5 format. '''
    _, class_values = get_label_info(class_dict_path)
    img2hdf5(img_dir, h5_path)
    label2hdf5(label_dir, h5_path, label_values=class_values)


def crop_imgs(image_dir, out_dir=None, crop_params=None, ifPad=True):
    ''' Slide crop images into small piece and save them. '''
    if out_dir is None:
        out_dir = image_dir

    image_names = sorted(os.listdir(image_dir))
    for name in tqdm(image_names):
        image = cv2.imread(image_dir+'/'+name, -1)
        # 附加操作
        # h, w = image.shape[:2]
        # image = cv2.resize(image, (int(w*0.7), int(h*0.7)))
        imgs, _ = slide_crop(image, crop_params, ifPad)
        for i in range(len(imgs)):
            save_name = rename_file(name, addstr="_%d" % (i+1))
            cv2.imwrite(out_dir+'/'+save_name, imgs[i], [1, 100])


if __name__ == '__main__':
    # main()
    # main_of_imgAndlabel_ops()
    # in_dir = '/home/tao/Code/Result/CVPR/测试集_真实/val_labels_visual'
    # in_dir = 'D:/Data_Lib/Other/Small/6/train_labels'
    # out_dir = '/home/tao/Code/Result/CVPR/测试集_真实/val_labels'
    # in_dir = '/home/tao/Data/RBDD/BackUp/Orignal_data/val_labels'
    # out_dir = '/home/tao/Data/RBDD/512/val_labels'
    # main_of_img_ops(in_dir, out_dir)

    # crop_imgs(in_dir, out_dir, [512, 512, 512])

    pass



'''
Tools collection of operations during training.
The icluded functions are:
    def train_log()
    def draw_loss_curve()
    def memory_watcher()
    def Accuracy Evaluation[serval functions]

Version 1.0  2018-04-02 22:44:32
    by QiJi Refence: https://github.com/GeorgeSeif/Semantic-Segmentation-Suite
Version 2.0  2018-10-29 09:10:41
    by QiJi
TODO:


'''
import datetime
import os
import sys

import numpy as np
from skimage import morphology
# from sklearn import metrics
from matplotlib import pyplot as plt


def train_log(X, f=None):
    ''' Print with time. To console or a file(f) '''
    time_stamp = datetime.datetime.now().strftime("[%d %H:%M:%S]")
    if not f:
        sys.stdout.write(time_stamp + " " + X)
        sys.stdout.flush()
    else:
        f.write(time_stamp + " " + X)


def draw_loss_curve(epochs, loss, path):
    ''' Paint loss curve '''
    fig = plt.figure(figsize=(12, 9))
    ax1 = fig.add_subplot(111)
    ax1.plot(range(epochs), loss)
    ax1.set_title("Average loss vs epochs")
    ax1.set_xlabel("Epoch")
    ax1.set_ylabel("Current loss")
    plt.savefig(path + '/loss_vs_epochs.png')
    plt.close(fig)


def memory_watcher():
    ''' Compute the memory usage, for debugging '''
    import psutil
    pid = os.getpid()
    py = psutil.Process(pid)
    memoryUse = py.memory_info()[0] / 2.**30  # Memory use in GB
    print('Memory usage in GBs:', memoryUse)


# **********************************************
# *********** Segmention Evaluation **************
# **********************************************
def fast_hist(label_true, label_pred, n_class):
    '''Computational confusion matrix.
    -------------------------------------------
    |          | p_cls_1 | p_cls_2 |   ....   |
    -------------------------------------------
    | gt_cls_1 |         |         |          |
    -------------------------------------------
    | gt_cls_2 |         |         |          |
    -------------------------------------------
    |   ....   |         |         |          |
    -------------------------------------------
    '''
    # mask = (label_true >= 0) & (label_true < n_class)
    if len(label_true.shape) > 1:
        label_true = label_true.flatten()
        label_pred = label_pred.flatten()
    hist = np.bincount(
        n_class * label_true.astype(int) + label_pred,
        minlength=n_class ** 2,
    ).reshape(n_class, n_class)
    return hist


def compute_global_accuracy(pred, label):
    '''
    Compute the average segmentation accuracy across all classes,
    Input [HW] or [HWC] label
    '''
    count_mat = pred == label
    return np.sum(count_mat) / np.prod(count_mat.shape)


def compute_class_accuracies(y_pred, y_true, num_classes):
    ''' Compute the class-specific segmentation accuracy '''
    w = y_true.shape[0]
    h = y_true.shape[1]
    flat_image = np.reshape(y_true, w * h)
    total = []
    for val in range(num_classes):
        total.append((flat_image == val).sum())

    count = [0.0] * num_classes
    for i in range(w):
        for j in range(h):
            if y_pred[i, j] == y_true[i, j]:
                count[int(y_pred[i, j])] = count[int(y_pred[i, j])] + 1.0
    # If there are no pixels from a certain class in the GT, it returns NAN
    # because of divide by zero, Replace the nans with a 1.0.
    accuracies = []
    for i in range(len(total)):
        if total[i] == 0:
            accuracies.append(1.0)
        else:
            accuracies.append(count[i] / total[i])

    return accuracies


def compute_class_iou(pred, gt, num_classes):
    '''
    Args:
        pred: Predict label [HW].
        gt: Ground truth label [HW].
    Return:
        （每一类的）intersection and union list.
    '''
    intersection = np.zeros(num_classes)
    union = np.zeros(num_classes)
    for i in range(num_classes):
        pred_i = pred == i
        label_i = gt == i
        intersection[i] = float(np.sum(np.logical_and(label_i, pred_i)))
        union[i] = float(np.sum(np.logical_or(label_i, pred_i)) + 1e-8)
    class_iou = intersection / union
    return class_iou


class runingScore(object):
    ''' Evaluation class '''
    def __init__(self, n_classes=2):
        self.n_classes = n_classes
        self.confusion_matrix = np.zeros((n_classes, n_classes), dtype=np.int64)

    def reset(self):
        ''' Reset confusion_matrix. '''
        self.confusion_matrix = np.zeros((self.n_classes, self.n_classes), dtype=np.int64)

    def update_all(self, label_trues, label_preds):
        ''' Add new pairs of predicted label and GT label to update the confusion_matrix.
        Note: Only suitable for segmentation
        '''
        for lt, lp in zip(label_trues, label_preds):
            self.confusion_matrix += fast_hist(lt, lp, self.n_classes)

    def print_score(self, score, mode=0):
        ''' Print the score dict.
        mode-0: print the final total scores
        mode-1: print per pair of data's scores
        '''
        str_score = ''
        for key in score:
            if 'Class' in key:
                value_str = ','.join('%.4f' % i for i in score[key])
            elif 'Hist' in key:
                continue
            else:
                value_str = '%.4f' % score[key]
            str_score += '%s,' % value_str if mode else key+': %s\n' % value_str
        str_score = str_score.strip(',').strip()  # discard the last suffix
        if mode == 0:
            print(str_score)

        return str_score


class classifyScore(runingScore):
    ''' Accuracy evaluation for classification(multi-class)'''
    def update(self, label_true, label_pred):
        '''Evaluate a new pair of predicted label and GT label,
        and update the confusion_matrix.'''
        hist = fast_hist(label_true, label_pred, self.n_classes)
        self.confusion_matrix += hist
        # return self.get_scores(hist)

    def get_scores(self, hist=None):
        """Returns accuracy score evaluation result.
            - Overall Acc
            - Class Acc
            - Mean Acc
        """
        hist = self.confusion_matrix if hist is None else hist
        # Overall accuracy
        acc_overall = np.diag(hist).sum() / (hist.sum()+1e-8)
        # Class accuracy
        acc_cls = np.diag(hist) / (hist.sum(axis=1)+1e-8)  # acc per class
        # Class average accuracy
        acc_cls_avg = np.nanmean(acc_cls)
        # Kappa
        n = hist.sum()
        p0 = hist.diagonal().sum()
        p1 = hist.sum(0)
        p2 = hist.sum(1)
        kappa = (n*p0-np.inner(p1, p2)) / (n*n - np.inner(p1, p2) + 1e-8)

        return (
            {
                "Hist": hist,  # 混淆矩阵
                "Kappa": kappa,
                "Overall Acc": acc_overall,
                "Class Acc": acc_cls,  # 类别精度
                "Mean Acc": acc_cls_avg,
            }  # Return as a dictionary
        )


class SegScore(runingScore):
    ''' Accuracy evaluation for semantic segmentation(multi-class)'''
    def update(self, label_true, label_pred):
        '''Evaluate a new pair of predicted label and GT label,
        and update the confusion_matrix.'''
        hist = fast_hist(label_true, label_pred, self.n_classes)
        self.confusion_matrix += hist
        return self.get_scores(hist)

    def get_scores(self, hist=None):
        """Use the confusion_matrix to do evaluation.
        Returns accuracy score evaluation result:
            - Overall Acc
            - Class Acc
            - Mean Acc
            - Class IoU
            - Mean IoU
            - FreqW Acc
        """
        hist = self.confusion_matrix if hist is None else hist
        # Overall accuracy
        acc_overall = np.diag(hist).sum() / hist.sum()
        # Class accuracy
        acc_cls = np.diag(hist) / hist.sum(axis=1)  # acc per class
        # Class average accuracy
        acc_cls_avg = np.nanmean(acc_cls)

        # IoU
        ep = np.ones((hist.shape[0])) * 1e-8
        cls_iu = np.diag(hist) / (hist.sum(axis=1) + hist.sum(axis=0) - np.diag(hist) + ep)
        mean_iu = np.nanmean(cls_iu)
        # Frequency Weighted IoU(FWIoU) 根据每个类出现的频率为其设置权重
        freq = hist.sum(axis=1) / hist.sum()
        fwavacc = (freq[freq > 0] * cls_iu[freq > 0]).sum()
        # cls_iu = dict(zip(range(self.n_classes), iu))

        return (
            {
                "Overall Acc": acc_overall,
                "Class Acc": acc_cls,  # 类别精度
                "Mean Acc": acc_cls_avg,
                "Class IoU": cls_iu,
                "Mean IoU": mean_iu,
                "FreqW Acc": fwavacc,
            }  # Return as a dictionary
        )


class RoadExtractionScore(runingScore):
    '''Accuracy evaluation for road extraction.
    Only two class: 0-bg, 1-road.
    '''

    def update(self, label_true, label_pred):
        '''Evaluate a new pair of predicted label and GT label,
        and update the confusion_matrix. '''
        hist = fast_hist(label_true, label_pred, self.n_classes)
        self.confusion_matrix += hist
        return self.get_scores(hist)

    def add(self, label_true, label_pred):
        '''Add a new pair of predicted label and GT label,
        update the confusion_matrix. '''
        hist = fast_hist(label_true, label_pred, self.n_classes)
        self.confusion_matrix += hist

    def get_scores(self, hist=None):
        """Returns accuracy score evaluation result.
            - 1. Precision{ TP / (TP+FP) }
            - 2. Recall{ TP / (TP+FN) }
            - 3. F1score
            - 4. Class IoU
            - 5. Mean IoU
            - 6. FreqW Acc
        """
        hist = self.confusion_matrix if hist is None else hist

        # Take class 1-road as postive class:
        TP = hist[1, 1]  # Ture Positive(road pixels are classified into road class)
        FN = hist[1, 0]  # False Negative(road pixels are classified into bg class)
        FP = hist[0, 1]  # False Positive(bg pixels are classified into road class)
        # TN = hist[0, 0]  # Ture Negative(bg pixels are classified into bg class)

        prec = TP / (TP + FP + 1e-8)  # Precision
        rec = TP / (TP + FN + 1e-8)  # Recall
        F1 = 2*TP / (2*TP + FP + FN + 1e-8)  # F1 Score

        # IoU (tested)
        cls_iu = np.diag(hist) / (hist.sum(axis=1) + hist.sum(axis=0) - np.diag(hist))
        mean_iu = np.nanmean(cls_iu)
        # Frequency Weighted IoU(FWIoU) 根据每个类出现的频率为其设置权重
        freq = hist.sum(axis=1) / hist.sum()
        fwavacc = (freq[freq > 0] * cls_iu[freq > 0]).sum()
        # cls_iu = dict(zip(range(self.n_classes), iu))

        return (
            {
                'Precision': prec,
                'Recall': rec,
                'F1score': F1,
                'Class IoU': cls_iu,
                'Mean IoU': mean_iu,
                'FreqW Acc': fwavacc,
            }  # Return as a dictionary
        )

    def keys(self):
        score_keys = [
            'Precision,Recall,F1score,Class IoU,Class IoU,Mean IoU,FreqW Acc'
            ]  # note 'Class IoU'
        return score_keys


class RelaxedRoadExtractionScore(runingScore):
    '''Relax Accuracy evaluation for road extraction.
    Only two class: 0-bg, 1-road.
    '''
    def __init__(self, rho=1):
        self.rho = rho*2 + 1
        self.confusion_matrix_p = np.zeros((2, 2), np.int64)  # For relaxed precision
        self.confusion_matrix_r = np.zeros((2, 2), np.int64)  # For relaxed recall

    def update(self, label_true, label_pred):
        '''Evaluate a new pair of predicted label and GT label,
        and update the confusion_matrix.'''
        if self.rho > 1:
            selem = morphology.square(self.rho, dtype=label_true.dtype)
            tp_label_true = morphology.dilation(label_true, selem)
            tp_label_pred = morphology.binary_dilation(label_pred, selem)
            hist1 = fast_hist(tp_label_true, label_pred, 2)
            hist2 = fast_hist(label_true, tp_label_pred, 2)
        else:
            hist = fast_hist(label_true, label_pred, 2)
            hist1, hist2 = hist, hist

        self.confusion_matrix_p += hist1
        self.confusion_matrix_r += hist2
        return self.get_scores(hist1, hist2)

    def add(self, label_true, label_pred):
        ''' Add new pairs of predicted label and GT label to update the confusion_matrix. '''
        if self.rho > 0:
            selem = morphology.square(self.rho, dtype=np.int64)
            tp_lt = morphology.binary_dilation(label_true, selem)
            tp_lp = morphology.binary_dilation(label_pred, selem)
            self.confusion_matrix_p += fast_hist(tp_lt, label_pred, 2)
            self.confusion_matrix_r += fast_hist(label_true, tp_lp, 2)
        else:
            hist = fast_hist(label_true, label_pred, 2)
            self.confusion_matrix_p += hist
            self.confusion_matrix_r += hist

    def get_scores(self, hist_p=None, hist_r=None):
        hist_p = self.confusion_matrix_p if hist_p is None else hist_p
        hist_r = self.confusion_matrix_r if hist_r is None else hist_r

        prec = hist_p[1, 1] / (hist_p[1, 1] + hist_p[0, 1] + 1e-8)  # Precision
        rec = hist_r[1, 1] / (hist_r[1, 1] + hist_r[1, 0] + 1e-8)  # Recall
        return (
            {
                "Precision": prec,
                "Recall": rec,
                # "F1score":
            }  # Return as a dictionary
        )

    def reset(self):
        ''' Reset confusion_matrixs. '''
        self.confusion_matrix_p = np.zeros((2, 2), dtype=np.int64)
        self.confusion_matrix_r = np.zeros((2, 2), dtype=np.int64)


# **********************************************
# ***********                     **************
# **********************************************
def test():
    # Test confusion_matrix
    # n_class = 2
    pre = np.zeros((10, 10), dtype=np.int32)
    gt = np.zeros((10, 10), dtype=np.int32)
    # # 0-BG; 1-Road
    pre[4:6, 4:6] = 1
    gt[4:6, 4:6] = 1  # 4 road piexls
    # print(gt)
    # print(pre)
    pre[4, 4] = 0  # 有1个原本是road 的被分为0-bg
    # pre[1:8, 1] = 1  # 有8个原本是bg 的被分为1-road
    # print(len(p.shape))
    # myscore = RelaxedRoadExtractionScore(rho=3)
    myscore = classifyScore(5)
    myscore.update(gt, pre)
    score = myscore.get_scores()
    myscore.print_score(score)
    # print('FP = ', hist[0, 1])
    # print('FN = ', hist[1, 0])
    # print('P = ', hist[1, 1] / (hist[1, 1]+hist[0, 1]))
    # print('R = ', hist[1, 1] / (hist[1, 1]+hist[1, 0]))
    # test = np.diag(hist)
    # print(test)


if __name__ == "__main__":
    test()
    pass



#!/usr/bin/env python
from __future__ import print_function
from setuptools import setup, find_packages
import sys

setup(
    name="qjdltools",
    version="0.1.1",
    author="Ji Qi",
    author_email="51978200@qq.com",
    description="Some simple tools for deep learning.",
    long_description=open("README.rst").read(),
    license="MIT",
    url="https://github.com/ErenTuring/dl_tools",
    packages=['qjdltools'],
    install_requires=[
        'h5py', 'matplotlib', 'opencv-python', 'tqdm',
        'scipy', 'scikit-image',
        ],
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
